// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2006 Shai Shalev-Shwartz and Yoram Singer.
//  All Rights Reserved.
//
// Classifier infrastructure and learning.
//
// Uses BLAS/ATLAS or an equivalent vector and matrix infrastructure.
//

#include <functional>

#include "classifier.h"
#include "bipartite_solver.h"

// Define of a very small number
#define MINISCULE_NUMBER 0.0000000001

// Define a stopping criterion accuracy
#define STOP_ACCURACY 0.00001

#define MAX_EPOCHS 1000

#define PDMODE(kname) ( (kname != "linear.kernel") ? true : false )

#define SUPPORT_SIZE(m) ( m > 50000 ? 50000 :  m )

// The basic constructor
Classifier::Classifier(int m, int n, int k,
                       std::string kernel_name, std::string d_method) :
  S(PDMODE(kernel_name) ? SUPPORT_SIZE(m) : 1, n),
  alpha(SUPPORT_SIZE(m), k),
  K(kernel_name),
  W(k,n),
  decomposer_(d_method),
  n_classes(k),
  dual_mode(PDMODE(kernel_name)),
  max_epochs(MAX_EPOCHS) {

  if ( dual_mode == false ) {
    W.zeros();
  }

  norm_of_w = 0.0;

  sum_gamma_alpha = 0.0;

  std::cerr << (dual_mode ? "Dual Mode\n" : "Primal Mode\n");
}



// The core prediction: multiply the support by an instance and then apply the
// kernel function to each element of the resulting product and finally
// multiply the result by the vector alpha.
infra::vector_view Classifier::Predict(infra::vector_view x) {

  if ( dual_mode == false ) {
    return W * x;
  } else {
    if ( S.Size() > 0 ) {
      return( K(S(),x) * alpha() );
    }  else { // If the support is empty return the zero vector.
      infra::vector predictions(n_classes);
      predictions.zeros();
      return predictions;
    }
  }
}



// The loss of a single bipartite graph
double Classifier::LossOfSubgraph(
    std::vector<int> &A, std::vector<int> &B,
    infra::vector_view &gamma, infra::vector_view &predictions) {

  double loss = 0.0;

  for ( int i = 0; i < A.size(); ++i ) {
    int a = A[i];
    for (uint j=0; j < B.size(); ++j) {
      int b = B[j];
      double cur_loss = gamma(a) - gamma(b) - predictions(a) + predictions(b);
      if ( cur_loss > loss )
        loss = cur_loss;
    }
  }

  return loss;
}



// Check whether there was error on a single graph:
//   Ignore the margin requirements and just verify that the predictions are
//   all in order
int Classifier::ErrorOfSubgraph(
    std::vector<int> &A, std::vector<int> &B,
    infra::vector_view &gamma, infra::vector_view &predictions) {

  for ( int i = 0; i < A.size(); ++i ) {
    int a = A[i];
    for (uint j=0; j < B.size(); ++j) {
      int b = B[j];
      if ( predictions(a) <= predictions(b) ) {
        return 1;
      }
    }
  }

  return 0;
}



// Internal loss calculation of all decompositions of an instance to subgraphs
double Classifier::LossFromDecomposition(infra::vector_view predictions,
                               infra::vector_view gamma,
                               decompositions::CostDecomposition &E) {

  // go over all the subgraphs and calculate the maximal hinge-loss
  double loss = 0.0;
  for ( int d = 0; d < E.Size(); ++d ) {
    double max_loss = LossOfSubgraph(E.A[d], E.B[d], gamma, predictions);
    loss += E.sigma[d] * max_loss;
  }
  return loss;
}



// Public loss calculation of an instance. Calls the internal version after
// generating the decomposition
double Classifier::Loss(infra::vector_view prediction,
                        infra::vector_view gamma) {

  // First generate the decomposition
  decompositions::CostDecomposition E;
  decomposer_.Generate(gamma, E);

  // Now call the graph's loss function
  return LossFromDecomposition(prediction, gamma, E);
}




// Internal error calculation of all decompositions of an instance to subgraphs.
// There is an error on a subgraph if its loss is greater than zero.
double Classifier::ErrorFromDecomposition(
    infra::vector_view predictions, infra::vector_view gamma,
    decompositions::CostDecomposition &E) {

  // go over all the subgraphs, calculate the hinge-loss and derive the error
  double error = 0.0;
  for ( int d = 0; d < E.Size(); ++d ) {
    int subgraph_err = ErrorOfSubgraph(E.A[d], E.B[d], gamma, predictions);
    if ( subgraph_err ) {
      error += E.sigma[d];
    }
  }
  return error;
}


// Public error calculation of an instance.
double Classifier::Error(infra::vector_view prediction,
                        infra::vector_view gamma) {

  // First generate the decomposition
  decompositions::CostDecomposition E;
  decomposer_.Generate(gamma, E);

  // Now call the graph's loss function
  return ErrorFromDecomposition(prediction, gamma, E);
}




double Classifier::Update(infra::vector_view x, infra::vector_view gamma,
                          int ind, double C) {

  // Save the old dual objective value for the end of the update
  double old_dual = this->DualObjective();

  // First, call the decomposition
  decompositions::CostDecomposition E;

  decomposer_.Generate(gamma, E);

  // Calculate the norm of x
  double norm_x = ( dual_mode ? K(x,x) : x * x );

  // predict using the previous ranker
  infra::vector_view old_predictions = Predict(x);

  // Get the previous value of alpha for the example and zero its alpha in case
  // it is in the support
  infra::vector prev_alpha(gamma.size());
  if ( alpha.Exist(ind) ) {
    prev_alpha = alpha.Get(ind);
  } else {
    prev_alpha.zeros();
  }

  // Calculate the predictions of U on x by eliminating the effect of the
  // current alpha for example ind. This amounts to:
  //            old_predictions - prev_alpha * norm_x
  infra::vector predictions = old_predictions - prev_alpha * norm_x;

  // allocate space for the current alpha
  infra::vector cur_alpha(gamma.size());
  cur_alpha.zeros();

  // If the decomposition is into a single graph then simply call the
  // Bipartite solve and we're done
  if ( E.Size() == 1 ) {

    // Calculate the new loss
    double max_loss = LossOfSubgraph(E.A[0], E.B[0], gamma, predictions);

    // Call the solver to find the best new alpha
    if (max_loss > 0.0) {
      BipartiteSolver(predictions, gamma, E.A[0], E.B[0], norm_x,
                      C*E.sigma[0], cur_alpha);
    }

  // In the case of multiple sub-graphs we need to iterate over the sub-graphs
  } else {
    // save the vector alpha
    infra::vector old_alpha = prev_alpha;

    // Set the "new gamma" to be gamma - predictions
    infra::vector_view new_gamma = gamma - predictions;

    // Set the predictions for the current instance to the zero vector
    predictions.zeros();

    // Create a matrix of alphas for the current instance
    infra::matrix all_alpha(E.Size(), gamma.size());
    all_alpha.zeros();

    // Perform multiple soft projections
    double new_dual = 0.0;
    for ( int iter = 0; iter < max_epochs; iter++ ) {

      // Check each sub-graph and perform multiprojection
      for ( int d = 0; d < E.Size(); ++d ) {

        // calculate the predictions with alpha of this subgraph set to zero
        predictions -= all_alpha.row(d) * norm_x;

        // and zero the alpha of this subgraph
        all_alpha.row(d).zeros();

        // Calculate the new loss
        double max_loss =
          LossOfSubgraph(E.A[d], E.B[d], new_gamma, predictions);

        // Update the predictions if the sub-graph is not predicted perfectly
        if (max_loss > 0.0) {
          BipartiteSolver(predictions, new_gamma, E.A[d], E.B[d],
                          norm_x, C * E.sigma[d], all_alpha.row(d));
          predictions += all_alpha.row(d) * norm_x;
        }
      }

      // The total alpha vector of the instance is the sum over d
      cur_alpha.zeros();
      for ( int i = 0; i < all_alpha.height(); ++i ) {
        cur_alpha += all_alpha.row(i);
      }

      // Stopping criterion: calculate dual (see Eq.(16) in the technical
      // report for further detail) and check for the increase in dual.
      double cur_dual =
        -0.5 * norm_x * cur_alpha.norm2() + cur_alpha * new_gamma;

      // If the dual did not increase much we can stop
      if ( cur_dual - new_dual <= STOP_ACCURACY ) {
        // std::cerr << "What inner? " << cur_dual << " " << new_dual << "\n";
        break;
      } else {
        new_dual = cur_dual;
      }
    }

    // calculate the dual for both old_alpha and cur_alpha
    double old_dual = -0.5 * norm_x * old_alpha.norm2() + old_alpha * new_gamma;

    new_dual = -0.5 * norm_x * cur_alpha.norm2() + cur_alpha * new_gamma;

    if ( new_dual < old_dual - MINISCULE_NUMBER ) {
      std::cerr << "What? " << old_dual << " " << new_dual << "\n";
      cur_alpha = old_alpha;
    }
  }

  // Calculate the difference between the previous alpha and current one
  infra::vector diff_alpha = cur_alpha - prev_alpha;

  // Update the norm of W. For efficency issues this is done in an indirect way
  // as follows. Let a_i,k be new value of alpha for example i and class k and
  // o_i,k be the old value of alpha. In addition denote the old vector for Wk
  // by Ok. Then,
  //
  // ||Wk||^2 = || sum_j alpha_j,k x_j ||^2 =
  //    || sum_{j<>i} alpha_j,k x_j + o_i,k x_i + (a_i,k - o_i,k) x_i ||^2 =
  //    || O_k + (a_i,k - o_i,k) x_i || =
  //    ||O_k||^2 + 2 (O_k dot x_i) (a_i,k-o_i,k) + ||x_i||^2 (a_i,k-o_i,k)^2
  //
  //        1                    2                            3
  //
  //  Summing over all k, the above equation consists of three terms:
  //  1) The old norm of W (squared)
  //  2) 2 x the innter product between the difference in alpha and O x_i which
  //  3) the norm of x_i times the norm of the difference vector ||a_i-o_i||^2
  //
  double old_norm_of_w = norm_of_w;
  norm_of_w +=
    2.0 * (old_predictions * diff_alpha) + (norm_x * diff_alpha.norm2());

  // update the sum of gamma * alpha
  sum_gamma_alpha += diff_alpha * gamma;

  // Insert or remove the instance from the support
  bool is_alpha_all_zero = true;
  for ( int i = 0; i < cur_alpha.size(); ++i ) {
    if ( fabs(cur_alpha(i)) > MINISCULE_NUMBER ) {
      is_alpha_all_zero = false;
    }
  }

  if ( alpha.Exist(ind) ) { // instance in support - check whether to remove
    if ( is_alpha_all_zero ) {
      alpha.Remove(ind);
      if ( dual_mode ) {
        S.Remove(ind);
      }
    } else
      alpha.Get(ind) = cur_alpha;
  } else {                 // instance not in support - check whether to add
    if ( !is_alpha_all_zero ) {
      if ( dual_mode ) {
        S.Insert(x,ind);
      }
      alpha.Insert(cur_alpha,ind);
    }
  }

  if ( !dual_mode ) {
    for ( int i = 0; i < W.height(); ++i ) {
      W.row(i) += diff_alpha(i) * x;
    }
  }

  // Finally return the increase in the dual
  return this->DualObjective() - old_dual;
}


double Classifier::NormSquared() {
  return norm_of_w;
}


double Classifier::DualObjective() {
  return -0.5 * norm_of_w + sum_gamma_alpha;
}
