// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2006 Shai Shalev-Shwartz and Yoram Singer.
//  All Rights Reserved.
//
// Classifier infrastructure and learning.
//
// Uses BLAS/ATLAS or an equivalent vector and matrix infrastructure.
//

#ifndef CLASSIFIER_H_
#define CLASSIFIER_H_

#include <vector>
#include <algorithm>

#include "kernels.h"
#include "infra/infra.h"
#include "support_set.h"
#include "decompositions.h"


class Classifier {
 public:

  Classifier(int m, int n, int k,
             std::string kernel, std::string decomposition);

  infra::vector_view Predict(infra::vector_view x);

  double Loss(infra::vector_view prediction,infra::vector_view gamma);

  double Error(infra::vector_view prediction,infra::vector_view gamma);

  double Update(infra::vector_view x, infra::vector_view gamma,
                int ind, double C);

  double NormSquared();

  double DualObjective();


  inline int NumberOfSupport() {
    return alpha.Size();
  }

  double norm_of_w;

  double sum_gamma_alpha;

 protected:

  // Flag for Primal/Dual mode: primal=0 dual=1 (default)
  bool dual_mode;

  // The classifier in Primal form (used only when dual_mode==false)
  infra::matrix W;

  // The support set of instances
  SupportSet S;

  // the coefficients of the support
  SupportSet alpha;

  kernels::KernelsBuilder K; // the kernel to use

  decompositions::DecompositionBuilder decomposer_; // the decomposition to use

  uint n_classes; // number of classes

  int max_epochs; // number of hildreth epochs for the mini problems

  // The maximal loss on a subgraph
  double LossOfSubgraph(std::vector<int> &A, std::vector<int> &B,
      infra::vector_view &gamma, infra::vector_view &predictions);

  // A 0-1 indicator for the error on a subgraph (1 == error occured)
  int ErrorOfSubgraph(std::vector<int> &A, std::vector<int> &B,
      infra::vector_view &gamma, infra::vector_view &predictions);

  // Calculate whole loss from for an instance given a decomposition
  double LossFromDecomposition(infra::vector_view prediction,
                               infra::vector_view gamma,
                               decompositions::CostDecomposition &E);

  // Calculate whole error from for an instance given a decomposition
  double ErrorFromDecomposition(infra::vector_view prediction,
                                infra::vector_view gamma,
                                decompositions::CostDecomposition &E);

};

#endif
