// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2006 Shai Shalev-Shwartz and Yoram Singer.
//  All Rights Reserved.
//
// Implementations of graph decomposition methods.
//
//

#include <vector>
#include "decompositions.h"

void decompositions::ZeroOneDecomposition::Generate(
    const infra::vector_view gamma, CostDecomposition &E)  {

  E.A.clear();
  E.B.clear();
  E.sigma.clear();

  // First, we would like to find all the vertices in the top most lattice
  // and all the rest of the vertices
  std::vector<int> top;
  std::vector<int> bottom;
  double max_gamma = gamma.max();
  for ( int i = 0; i < gamma.size(); ++i ) {
    if ( gamma(i) == max_gamma )
      top.push_back(i);
    else
      bottom.push_back(i);
  }

  // We have only a single subgraph
  E.A.push_back(top);
  E.B.push_back(bottom);
  E.sigma.push_back(1.0);
}



void decompositions::HalfHalfDecomposition::Generate(
    const infra::vector_view gamma, CostDecomposition &E)  {
  E.A.clear();
  E.B.clear();
  E.sigma.clear();

  std::vector<double> sorted_gamma(gamma.size());

  for ( int i = 0; i < gamma.size(); ++i )
    sorted_gamma[i] = gamma[i];

  sort(sorted_gamma.begin(), sorted_gamma.end(), std::greater<double>());

  int mid = gamma.size() / 2;
  double split_gamma = sorted_gamma[mid];

  std::vector<int> top;
  std::vector<int> bottom;

  for ( int i = 0; i < gamma.size(); ++i ) {
    if ( gamma[i] <= split_gamma ) {
      bottom.push_back(i);
    } else {
      top.push_back(i);
    }
  }

  // We have only single subgraph here
  E.A.push_back(top);
  E.B.push_back(bottom);
  E.sigma.push_back(1.0);
}



void decompositions::AllPairsDecomposition::Generate(
    const infra::vector_view gamma,CostDecomposition & E)  {

  E.A.clear();
  E.B.clear();
  E.sigma.clear();

  int d = 0; // the number of subgraphs

  // Check all pairs and add pairs that are in the correct relation
  for ( int i = 0; i < gamma.size() ; ++i ) {
    for ( int j = 0; j < gamma.size() ; ++j ) {
      std::vector<int> local_A(1), local_B(1);

      if ( gamma[i] > gamma[j] ) {
        local_A[0] = i;
        local_B[0] = j;
        E.A.push_back(local_A);
        E.B.push_back(local_B);
        ++d;
      }
    }
  }

  E.sigma.resize(d);
  double local_sigma = 1.0 / ((double) d);
  for ( int i = 0; i < d; ++i ) {
    E.sigma[i] = local_sigma;
  }
}


void decompositions::DominationDecomposition::Generate(
    const infra::vector_view gamma, CostDecomposition &E)  {

  E.A.clear();
  E.B.clear();
  E.sigma.clear();

  int d = 0; // Number of decompositions

  // Check all label and create a list of the nodes each label dominates
  for ( int i = 0; i < gamma.size() ; ++i ) {
    std::vector<int> local_A(1), local_B;

    local_A[0] = i;
    local_B.clear();

    for ( int j = 0; j < gamma.size() ; ++j ) {
      if ( gamma[i] > gamma[j] ) {
        local_B.push_back(j);
      }
    }

    if ( local_B.size() > 0 ) {
      E.A.push_back(local_A);
      E.B.push_back(local_B);
      ++d;
    }
  }

  E.sigma.resize(d);
  double local_sigma = 1.0 / ((double) d);
  for ( int i = 0; i < d; ++i ) {
    E.sigma[i] = local_sigma;
  }
}
