// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2006 Shai Shalev-Shwartz and Yoram Singer.
//  All Rights Reserved.
//
// Implementation of graph decomposition methods:
//    break a preference function into two disjoint sets A and B
//
//

#ifndef DECOMPOSITION_H_
#define DECOMPOSITION_H_

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstdlib>

#include "infra/infra.h"

namespace decompositions {

  // A decomposition is a partition into multiple disjoint bipartite graphs
  class CostDecomposition {

  public:

    CostDecomposition() : A() , B(), sigma() { }

    // Return number of actual elements in the top set of the bipartite graph
    int Size() {
      return A.size();
    }

    // A[d][i] designates the i'th label in the set A of the
    // d'th bipartite graph
    std::vector< std::vector<int> > A;

    // B[d][i] designates the i'th label in the set B of the
    // d'th bipartite graph
    std::vector< std::vector<int> > B;

    // sigma[d] is the weight of the d'th graph
    std::vector< double > sigma;

  };

  // The decomposition opeartor:
  //    gets a vector of preferences gamma and returns a cost decomposition.
  class BaseDecomposition {

    public:
      // gamma: target ranking    E - resulting decomposition
      virtual void Generate(
          const infra::vector_view gamma, CostDecomposition &E) { }

      virtual ~BaseDecomposition() { }

    protected:
      BaseDecomposition( ) { }

  };


  // The zero-one decomposition is one big bipartite graph
  class ZeroOneDecomposition : public BaseDecomposition {
    public:
      ZeroOneDecomposition( ) { }
      void Generate(const infra::vector_view gamma, CostDecomposition &E);
  };


  // The HalfHalf decomposition generates one graph which compares the top
  // ranked labels to the rest of the labels
  class HalfHalfDecomposition : public BaseDecomposition {
    public:
      HalfHalfDecomposition( ) { }
      void Generate(const infra::vector_view gamma, CostDecomposition &E);
  };


  // Domination decompostion has multiple graphs, one for each node/labels
  // which dominates other labels.
  class DominationDecomposition : public BaseDecomposition {
    public:
      DominationDecomposition() { }
      void Generate(const infra::vector_view gamma, CostDecomposition &E);
  };


  // All-pairs generates graphs each of which has just two niodes
  class AllPairsDecomposition : public BaseDecomposition {
    public:
      AllPairsDecomposition()  { }
      void Generate(const infra::vector_view gamma, CostDecomposition &E);
  };


  // Interface to generate decomposition from run-time parameters
  class DecompositionBuilder {
    public:

      enum DecompositionTypes {
        ZeroOne     = 0,
        AllPairs    = 1,
        Domination  = 2,
        HalfHalf    = 3
      };

      // Construct a decompostion from a name
      DecompositionBuilder(std::string method_name) {
        if (method_name == "ZeroOne") {
          decomposition_method_ = new ZeroOneDecomposition;
          decomposition_code_ = ZeroOne;
        } else if (method_name == "AllPairs") {
          decomposition_method_ = new AllPairsDecomposition;
          decomposition_code_ = AllPairs;
        } else if (method_name == "Domination") {
          decomposition_method_ = new DominationDecomposition;
          decomposition_code_ = Domination;
        } else if (method_name == "HalfHalf") {
          decomposition_method_ = new HalfHalfDecomposition;
          decomposition_code_ = HalfHalf;
        } else {
          std::cerr << "Unknown decomposition method (" <<
            method_name << ")" << std::endl;
          exit(-1);
        }
      }

      // Construct a decompoisition method from a file
      DecompositionBuilder(FILE *fd) {
        unsigned int code;

        infra::load_binary(fd, code);

        switch (code) {
          case ZeroOne:
            decomposition_method_ = new ZeroOneDecomposition;
            break;

          case AllPairs:
            decomposition_method_ = new AllPairsDecomposition;
            break;

          case Domination:
            decomposition_method_ = new DominationDecomposition;
            break;

          case HalfHalf:
            decomposition_method_ = new HalfHalfDecomposition;
            break;

          default:
            std::cerr << "Unknown decomposition code (" <<
              code << ")" << std::endl;
            exit(-1);
        }

        decomposition_code_ = code;
      }

      void Generate(const infra::vector_view gamma, CostDecomposition &E) {
        decomposition_method_->Generate(gamma, E);
      }

      void Save(FILE *fd) {
        unsigned int code = decomposition_code_;
        infra::save_binary(fd, code);
      }

      ~DecompositionBuilder() {
        delete decomposition_method_;
      }

    protected:
      // Pointer to the actual decomposition methods used
      BaseDecomposition *decomposition_method_;

      // The code of the decomposition used
      int decomposition_code_;
  };

};

#endif // DECOMPOSITION_H_
