// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2006 Shai Shalev-Shwartz and Yoram Singer.
//  All Rights Reserved.
//
// Defintions and classes for various kernels.
//
// Uses BLAS/ATLAS or an equivalent vector and matrix infrastructure.
//

#ifndef KERNELS_H_
#define KERNELS_H_

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstdlib>

#include "infra/infra.h"

namespace kernels {

  // Pure virtual class for kernels
  class BaseKernel {

    public:

      // The "classical" instance-instance operator
      virtual double operator() (
          const infra::vector_view v1, const infra::vector_view v2) const = 0;

      // The Gram matrix M1*M2' calculated with respect to the kernel
      // operator. Coordiante [i,j] of the outcome is the i'th row of M1
      // multplied by the j'th row of M2
      virtual infra::matrix_view operator() (
          const infra::matrix_view M1, const infra::matrix_view M2) const = 0;

      // Vector of inner products of each row in M and v via a kernel operator
      virtual infra::vector_view operator() (
          const infra::matrix_view M, const infra::vector_view v) const = 0;

      void save_binary(FILE *fd) {
        param_.save_binary(fd);
      }

      virtual ~BaseKernel() { }

    protected:

      BaseKernel(infra::vector_view v) : param_(v) { }

      // The kernel specific parameters.
      infra::vector param_;
  };



  // Linear kernel:
  //    standard inner-product with an additional affine transofmration
  //    K(x,y) = a * (<x,y> + b) where v(0)=a and v(1) = b
  class linear : public BaseKernel {

    public:

      linear(infra::vector_view v) : BaseKernel(v) {
        bool wrong_parameters = false;
        if ( v.size() != 2 ) {
          wrong_parameters = true;
        } else if ( v(0) <= 0 || v(1) < 0 ) {
          wrong_parameters = true;
        }
        if (wrong_parameters) {
          std::cerr << "Linear kernel with wrong parameters: " << v << "\n";
          exit(-1);
        }

        _b = param_(1); _s = param_(0);
      }

      inline double operator() (const infra::vector_view v1,
                                const infra::vector_view v2) const;

      inline infra::matrix_view operator() (const infra::matrix_view M1,
                                            const infra::matrix_view M2) const;

      inline infra::vector_view operator() (const infra::matrix_view M,
                                            const infra::vector_view v) const;

      ~linear() { }

    protected:
      double _b;
      double _s;
  };

  // K(x,y) = s * (<x,y> + b)^d = v(0) * (<x,y> + v(1))^(v(2))
  class polynomial : public BaseKernel {
    public:

      polynomial(infra::vector_view v) : BaseKernel(v) {
        bool wrong_parameters = false;
        if ( v.size() != 3 ) {
          wrong_parameters = true;
        } else if ( v(0) <= 0 || v(1) < 0 || v(2) <= 0 ) {
          wrong_parameters = true;
        }
        if (wrong_parameters) {
            std::cerr << "Poly kernel with wrong parameters: " << v << "\n";
            exit(-1);
        }
        _d = v(2); _b = v(1); _s = v(0);
      }

      inline double operator() (const infra::vector_view v1,
                                const infra::vector_view v2) const;

      inline infra::matrix_view operator() (const infra::matrix_view M1,
                                            const infra::matrix_view M2) const;

      inline infra::vector_view operator() (const infra::matrix_view M,
                                            const infra::vector_view v) const;

      ~polynomial() { }

    protected:

      double _d, _b, _s;
  };

  // Constructor K(x,y) = v(0) * exp( -||x-y||^2 / 2 v(1))
  class gaussian : public BaseKernel {

  public:

    gaussian(infra::vector_view v) : BaseKernel(v) {
      bool wrong_parameters = false;
      if ( v.size() != 2 ) {
        wrong_parameters = true;
      } else if ( v(0) <= 0 || v(1) <= 0 ) {
        wrong_parameters = true;
      }
      if (wrong_parameters) {
        std::cerr << "RBF kernel with wrong parameters: " << v << "\n";
        exit(-1);
      }
      _s = log(param_(0)); _sigma2  = param_(1);
    }

    inline double operator() (const infra::vector_view v1,
                              const infra::vector_view v2) const;

    inline infra::matrix_view operator() (const infra::matrix_view M1,
                                          const infra::matrix_view M2) const;

    inline infra::vector_view operator() (const infra::matrix_view M,
                                          const infra::vector_view v) const;

    ~gaussian() { }

    protected:
      double _sigma2;
      double _s;
  };



  /* A simple kernel builder.
     For example: create a file linear.kernel which contains:
      linear
      2 1 0

     The code should include the following file:
        kernels::KernelsBuilder K(kernel_filename);

     Possible usages:
        double x_norm = K(x,x);
        infra::vector_view v = K(M,x);
  */

  class KernelsBuilder {

    public:

      enum KernelsTypes { LINEAR = 0 , POLYNOMIAL = 1, GAUSSIAN = 2 };

      KernelsBuilder(std::string file_name) : K(NULL) {
        if (file_name == "default") {
          infra::vector v(2); v(0) = 1.0; v(1) = 1.0;
          K = new gaussian(v);
          kernel_code = GAUSSIAN;
        } else {
          std::ifstream kernel_file(file_name.c_str());
          // read the name of the kernel
          std::string s;
          kernel_file >> s;

          // read the infra::vector of paramters
          infra::vector v(kernel_file);

          // construct the kernel
          if (s == "linear") {
            K = new linear(v);
            kernel_code = LINEAR;
          } else if (s == "polynomial") {
            K = new polynomial(v);
            kernel_code = POLYNOMIAL;
          } else if (s == "gaussian") {
            K = new gaussian(v);
            kernel_code = GAUSSIAN;
          } else {
            std::cerr << "unknown kernel name " << s << std::endl;
            exit(-1);
          }
        }
      }


      KernelsBuilder(FILE *fd) : K(NULL) {
        // read the code for the kernel
        infra::load_binary(fd, kernel_code);

        // read the infra::vector of parameters
        infra::vector v(fd);

        if (kernel_code == LINEAR) {
          K = new linear(v);
        } else if (kernel_code == POLYNOMIAL) {
          K = new polynomial(v);
        } else if (kernel_code == GAUSSIAN) {
          K = new gaussian(v);
        } else {
          std::cerr << kernel_code << " is an unknown." << std::endl;
          exit(-1);
        }
      }


    double operator() (const infra::vector_view v1,
                       const infra::vector_view v2) const {
      return (*K)(v1, v2);
    }

    infra::matrix_view operator() (const infra::matrix_view M1,
                                   const infra::matrix_view M2) const {
      return (*K)(M1, M2);
    }

    infra::vector_view operator() (const infra::matrix_view M,
                                   const infra::vector_view v) const {
      return (*K)(M, v);
    }

    ~KernelsBuilder() {
      delete K;
    }

    protected:
      BaseKernel *K; // Actual kernel used
      short kernel_code;
  };

};


// -------------------------------------------------------------------
//   Various Implementations of Kernel functions
// -------------------------------------------------------------------


inline double kernels::linear::operator() (
  const infra::vector_view v1, const infra::vector_view v2) const {
  return ((v1 * v2) + _b) * _s;
}

inline infra::matrix_view kernels::linear::operator() (
  const infra::matrix_view M1, const infra::matrix_view M2) const {

  infra::matrix outcome(M1.height(), M2.height());
  infra::prod_t(M1, M2, outcome);
  if ( _b != 0.0 ) {
    outcome += _b;
  }
  outcome *= _s;
  return outcome;
}


inline infra::vector_view kernels::linear::operator() (
  const infra::matrix_view M, const infra::vector_view v) const {

  infra::vector_view outcome = M * v;
  if ( _b != 0.0 ) {
    outcome += _b;
  }
  outcome *= _s;
  return outcome;
}


inline double kernels::polynomial::operator() (
  const infra::vector_view v1, const infra::vector_view v2) const {
  return pow((v1 * v2) + _b,_d) * _s;
}


inline infra::matrix_view kernels::polynomial::operator() (
  const infra::matrix_view M1, const infra::matrix_view M2) const {

  infra::matrix outcome(M1.height(), M2.height());
  infra::prod_t(M1, M2, outcome);
  if ( _b != 0.0 ) {
    outcome += _b;
  }
  outcome.pow(_d);
  outcome *= _s;
  return outcome;
}


inline infra::vector_view kernels::polynomial::operator() (
  const infra::matrix_view M, const infra::vector_view v) const {

  infra::vector_view outcome = M * v;
  if (_b != 0.0) {
    outcome += _b;
  }
  outcome.pow(_d);
  outcome *= _s;
  return outcome;
}


inline double kernels::gaussian::operator() (
  const infra::vector_view v1, const infra::vector_view v2) const {

  return exp(_s - infra::dist2(v1,v2) / 2.0 / _sigma2);
}

inline infra::matrix_view kernels::gaussian::operator() (
  const infra::matrix_view M1, const infra::matrix_view M2) const {

  infra::matrix outcome(M1.height(), M2.height());
  infra::prod_t(M1, M2, outcome);
  infra::vector M1_norms2(M1.height());
  infra::vector M2_norms2(M2.height());

  for ( int i = 0; i < M1.height(); ++i ) {
    M1_norms2(i) = M1.row(i) * M1.row(i) / 2.0;
  }

  for ( int i = 0; i < M2.height(); ++i ) {
    M2_norms2(i) = M2.row(i) * M2.row(i) / 2.0;
  }

  for( int j = 0; j < outcome.width(); ++j) {
    for(unsigned long i=0; i<outcome.height(); ++i) {
      outcome(i,j) =
        exp(_s - (M1_norms2(i) + M2_norms2(j) - outcome(i,j)) / _sigma2);
    }
  }

  return outcome;
}

inline infra::vector_view kernels::gaussian::operator() (
  const infra::matrix_view M, const infra::vector_view v) const {

  infra::vector_view outcome = M * v;
  infra::vector M_norms2(M.height());
  for ( int i = 0; i < M.height(); ++i ) {
    M_norms2(i) = M.row(i) * M.row(i) / 2.0;
  }

  double v_norm2 = v * v / 2.0;
  for( int i = 0; i < outcome.size(); ++i ) {
    outcome(i) = exp(_s - (M_norms2(i) + v_norm2 - outcome(i)) / _sigma2);
  }

  return outcome;
}

#endif
