// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2006 Shai Shalev-Shwartz and Yoram Singer.
//  All Rights Reserved.
//
// Main for the SOPOPO algorithm.
//

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstdlib>

#include "kernels.h"
#include "classifier.h"
#include "dataset.h"

// Define a duality gap accuracy
#define DUALITY_GAP_ACCURACY 0.01

// print usage massage and exit
void usage(char *prog_name) {
    std::cerr << "USAGE:\n\t" << prog_name << " [options] instances_file labels_file\n";
    std::cerr << "\nDESCRIPTION:\n"
          << "\tImplement the sopopo algorithm for label ranking by Shalev-Shwartz and Singer\n"
          << "\tThe instances and labels file should be in infra format (see README file)\n";
    std::cerr << "\nOPTIONS:\n"
          << "\t-C number\n\t\tThe parameter C of SVM (default: 1.)\n"
          << "\t-e number\n\t\tThe maximal number of epochs (default: 3)\n"
          << "\t-d string\n\t\tone of the following Decomposition methods:\n"
          << "\t\tZeroOne AllPairs Domination HalfHalf (default: ZeroOne)\n"
          << "\t-k string\n\t\tName of a kernel description file (default: linear.kernel)\n"
          << "\t\tSee kernels.h for more details\n"
          << "\t-v string1 string2\n"
          << "\t\tNames of validation instances file and lables file (default: no validation)\n";
    exit(EXIT_FAILURE);
}


int main(int argc, char** argv) {

  if ( argc < 3 ) {
    usage(argv[0]);
  }

  // Parse command line
  std::string instance_filename(argv[argc-2]);
  std::string label_filename(argv[argc-1]);
  double C = 1;
  int epochs = 3;
  std::string decomposition_method("ZeroOne");
  std::string kernel_filename("linear.kernel");
  bool validate = false;
  std::string v_instance_filename;
  std::string v_label_filename;
  int i=1;
  while (i<argc-2) {
    if (argv[i][0] != '-')
      usage(argv[0]);

    switch( argv[i][1] ) {
    case 'C': case 'c':
      C = atof(argv[i+1]);
      i += 2;
      break;
    case 'e':
      epochs = atoi(argv[i+1]);
      i += 2;
      break;
    case 'd':
      decomposition_method = argv[i+1];
      i += 2;
      break;
    case 'k':
      kernel_filename = argv[i+1];
      i += 2;
      break;
    case 'v':
      validate = true;
      v_instance_filename = argv[i+1];
      v_label_filename = argv[i+2];
      i += 3;
      break;
    default:
      usage(argv[0]);
    }

  }
  if (i != argc-2)
    usage(argv[0]);


  // Open training dataset
  DataSet data(instance_filename, label_filename);


  int m; // number of example
  int n; // dimension of each instance
  int n_classes; // number of classes

  data.open(m, n, n_classes);

  std::cerr << "m = " << m << " n = " << n << " n_classes = " <<  n_classes << "\n";

  infra::vector instance(n);

  infra::vector gamma(n_classes);

  // Construct a classifier
  Classifier w(m, n, n_classes, kernel_filename, decomposition_method);

  // variable for stopping criterion
  double cur_dual = 0.0;


  // Cycle through the data
  for ( int iter = 0; iter < epochs; ++iter ) {

    // reopen data
    data.reopen();

     // go over the examples and update
    for ( int j = 0; j < m; ++j ) {

      // read next instance and label
      data.next(instance, gamma);

      // update
      w.Update(instance, gamma, j, C);

      double new_dual = w.DualObjective();

      // std::cout << "Dual objective = " << new_dual << std::endl;

      if ( cur_dual - new_dual > 0.00000000001 ) {
        std::cerr << "Dual decreased. Aborting\n";
        exit(EXIT_FAILURE);
      }

      cur_dual = new_dual;

#ifdef XTRADBG
      if ( (j+1) % 10000 == 0 ) {
        fprintf(stderr, "%5d %5d\n", j+1, w.NumberOfSupport());
      }
#endif
    }

#ifdef XTRADBG
    fprintf(stderr, "\n");
#endif


    // Calculate the primal value
    double loss = 0.0, error = 0.0;
    data.reopen();
    for ( int j = 0; j < m; ++j ) {
      data.next(instance, gamma);
      infra::vector prediction(w.Predict(instance));
      double cur_loss = w.Loss(prediction, gamma);
      loss += cur_loss;
      double cur_error = w.Error(prediction, gamma);
      error += cur_error;
    }

    double primal = 0.5 * w.NormSquared() + C * loss;

    error /= m;

    fprintf(stderr, "%3d: Dual: %g  Primal: %g  Gap: %g",
            iter, cur_dual, primal, primal - cur_dual);
    fprintf(stderr, "  Norm^2: %g  Loss:  %g  Error: %g Support: %d\n",
            w.NormSquared(), loss, error, w.NumberOfSupport());

    if (validate) {

      DataSet validation(v_instance_filename, v_label_filename);
      int vm; // validation - number of example
      int vn; // validation - dimension of each instance
      int vn_classes; // validation - number of classes

      validation.open(vm, vn, vn_classes);

      std::cerr << "vm = " <<vm << " vn = " <<vn << " vn_classes = " << vn_classes << "\n";

      // Caluclate loss and error on the validation set
      double vloss = 0.0, verror = 0.0;
      validation.reopen();
      for ( int j = 0; j < vm; ++j ) {
        validation.next(instance, gamma);
        infra::vector prediction(w.Predict(instance));
        double cur_loss = w.Loss(prediction, gamma);
        vloss += cur_loss;
        double cur_error = w.Error(prediction, gamma);
        verror += cur_error;
      }

      verror /= vm;
      vloss /= vm;

      fprintf(stderr, "     Validation - Loss: %g Error: %g\n",
          vloss, verror);
    }


    // check duality gap
    if (primal - cur_dual <= DUALITY_GAP_ACCURACY) {
      std::cout << "Success." << std::endl;
      break;
    }
  }
}
