// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2006 Shai Shalev-Shwartz and Yoram Singer.
//  All Rights Reserved.
//
// Implementation of a support set. Assumes a fixed size for the support and
// a variable size for the alpha's.
//
// Uses BLAS/ATLAS or an equivalent vector and matrix infrastructure.
//


#include "support_set.h"

infra::matrix_view SupportSet::operator() () {
  return support_.submatrix(0, 0, support_count_, support_.width());
}


infra::vector_view SupportSet::Get(int ind) {
  int pos = index_.find(ind)->second;
  return support_.row(pos);
}



void SupportSet::Insert(infra::vector_view &instance, int ind) {

  // First check is the example is already in the support and do nothing
  // if it is indeed in.
  if ( index_.find(ind) != index_.end() )
    return;

  // Check if we did not exhaust the support.
  if ( support_count_ >= support_.height() ) {
    std::cerr << "Reached the maximum size of the support.\n";
    exit(-1);
  }

  // Finally, it is safe to insert the instance to the support -- just do it
  support_.row(support_count_) = instance;
  index_[ind] = support_count_;
  inverse_index_[support_count_] = ind;

  ++support_count_;
}



void SupportSet::Remove(int ind) {

  // First check is the example is in the support. If it is not, nothing
  // needs to be done
  if ( index_.find(ind) == index_.end() ) {
    return;
  }

  int pos = index_.find(ind)->second;

  // Swap the last instance in the support with the instance to be removed
  if ( pos != support_count_-1 ) {

    // Move the last row to the place where the removed row is
    support_.row(pos) = support_.row(support_count_-1);

    // Change the indices for the removed and moved rows
    int moved_ind = inverse_index_[support_count_-1];
    inverse_index_[pos] = moved_ind;
    index_[moved_ind] = pos;
  }

  // Erase from the index
  index_.erase(ind);

  --support_count_;
}



bool SupportSet::Exist(int ind) {
  return ( index_.find(ind) != index_.end() );
}



void SupportSet::Save(FILE *fd) {

  std::cerr << "Save of support sets is not implemented\n";
  exit(-1);

#ifdef notdef
  // Write the relevant portion of the support
  support_.submatrix(0, 0, support_count_,support_.width()).save_binary(fd);

  // write the number of examples which is also the size of the index
  unsigned int n_examples = index_.size();
  infra::save_binary(fd, n_examples);

  // write Index_ and Inverse_Index_
  for ( int i = 0; i < index_.size(); ++i ) {
    infra::save_binary(fd, (unsigned int) index_[i]);
  }

  for ( int i = 0; i < inverse_index_.size(); ++i ) {
    infra::save_binary(fd, (unsigned int) inverse_index_[i]);
  }
#endif

}



SupportSet::SupportSet(FILE * fd) : support_(fd)
{
  std::cerr << "Construction of support sets from file is not implemented\n";
  exit(-1);

#ifdef notdef
  support_count_ = support_.height();

  // load the number of examples == size of index
  unsigned int n_examples;
  infra::load_binary(fd, n_examples);

  index_.resize(n_examples);
  for ( int i = 0; i < n_examples; ++i ) {
    unsigned int tmp;
    infra::load_binary(fd, tmp);
    index_[i] = tmp;
  }

  inverse_index_.resize(support_count_);
  for ( int i = 0; i < inverse_index_.size(); ++i ) {
    unsigned int tmp;
    infra::load_binary(fd, tmp);
    inverse_index_[i] = tmp;
  }
#endif
}
